﻿using System;
using System.Collections.Generic;
using UniRx;
using UnityEngine;

namespace Magify
{
    public class FeaturesProvider : IDisposable
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(MagifyService.LogScope);

        public delegate bool TryGetFunc<T>(string featureName, out T result);

        private class Property<T>
        {
            private readonly Dictionary<string, ReactiveProperty<T>> _properties = new();

            private readonly TryGetFunc<T> _getter;

            public Property(TryGetFunc<T> getter)
            {
                _getter = getter;
            }

            internal void UpdateValues()
            {
                foreach (var property in _properties)
                {
                    var featureName = property.Key;
                    if (_getter(featureName, out var value))
                    {
                        property.Value.Value = value;
                    }
                }
            }

            public IReadOnlyReactiveProperty<T> this[string featureName, T defaultValue = default]
            {
                get
                {
                    if (!_properties.TryGetValue(featureName, out var prop))
                    {
                        if (!_getter(featureName, out var value))
                        {
                            value = defaultValue;
                        }
                        prop = new ReactiveProperty<T>(value);
                        _properties[featureName] = prop;
                    }

                    return prop;
                }
            }
        }

        private readonly Property<bool> _booleans = new(MagifyManager.Features.TryGetBool);
        private readonly Property<double> _numbers = new(MagifyManager.Features.TryGetNumber);
        private readonly Property<string> _strings = new(MagifyManager.Features.TryGetString);

        public event Action OnFeaturesUpdated
        {
            add => MagifyManager.Features.OnFeaturesUpdated += value;
            remove => MagifyManager.Features.OnFeaturesUpdated -= value;
        }

        public bool TryGetBool(string featureName, out bool result) => MagifyManager.Features.TryGetBool(featureName, out result);
        public bool GetBool(string featureName) => MagifyManager.Features.GetBool(featureName);
        public IReadOnlyReactiveProperty<bool> GetReactiveBool(string featureName, bool defaultValue = default) => _booleans[featureName, defaultValue];

        public bool TryGetNumber(string featureName, out double result) => MagifyManager.Features.TryGetNumber(featureName, out result);
        public double GetNumber(string featureName) => MagifyManager.Features.GetNumber(featureName);
        public IReadOnlyReactiveProperty<double> GetReactiveNumber(string featureName, double defaultValue = default) => _numbers[featureName, defaultValue];

        public bool TryGetString(string featureName, out string result) => MagifyManager.Features.TryGetString(featureName, out result);
        public string GetString(string featureName) => MagifyManager.Features.GetString(featureName);
        public IReadOnlyReactiveProperty<string> GetReactiveString(string featureName, string defaultValue = default) => _strings[featureName, defaultValue];

        public FeaturesProvider()
        {
            MagifyManager.Features.OnFeaturesUpdated += FeaturesUpdatedHandler;
        }

        void IDisposable.Dispose()
        {
            MagifyManager.Features.OnFeaturesUpdated -= FeaturesUpdatedHandler;
        }

        private void FeaturesUpdatedHandler()
        {
            _logger.Log("Update reactive features");
            _booleans.UpdateValues();
            _numbers.UpdateValues();
            _strings.UpdateValues();
        }
    }
}